@@ -1,5 +1,5 @@
 /*
- *  R : A Computer Langage for Statistical Data Analysis
+ *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -17,12 +17,6 @@
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-	/* NOTE ON FACTORS */
-	/* For us, c(factor1, factor2) is legal only if factor1 */
-	/* and factor2 are either both ordered or both unordered */
-	/* and have the same levels.  Any attempt at combining */
-	/* factors with anything else is stamped on. */
-
 #include "Defn.h"
 
 static SEXP cbind(SEXP, SEXP, SEXPTYPE);
@@ -68,14 +62,9 @@ static void answertype(SEXP x, int recurse, int usenames)
 				ans_flags |= 1;
 				ans_length += LENGTH(CAR(t));
 				break;
-			case FACTSXP:
-				ans_flags |= 2;
-				ans_length += LENGTH(CAR(t));
-				break;
-			case ORDSXP:
-				ans_flags |= 4;
-				ans_length += LENGTH(CAR(t));
-				break;
+
+			/* factors here previously */
+
 			case INTSXP:
 				ans_flags |= 8;
 				ans_length += LENGTH(CAR(t));
@@ -84,12 +73,10 @@ static void answertype(SEXP x, int recurse, int usenames)
 				ans_flags |= 16;
 				ans_length += LENGTH(CAR(t));
 				break;
-#ifdef COMPLEX_DATA
 			case CPLXSXP:
 				ans_flags |= 32;
 				ans_length += LENGTH(CAR(t));
 				break;
-#endif
 			case STRSXP:
 				ans_flags |= 64;
 				ans_length += LENGTH(CAR(t));
@@ -128,16 +115,6 @@ static void listanswer(SEXP x, int recurse)
 			ans_ptr = CDR(ans_ptr);
 		}
 		break;
-	case FACTSXP:
-	case ORDSXP:
-		for(i=0 ; i<LENGTH(x) ; i++) {
-			CAR(ans_ptr) = allocVector(TYPEOF(x), 1);
-			LEVELS(CAR(ans_ptr)) = LEVELS(x);
-			setAttrib(CAR(ans_ptr), R_LevelsSymbol, getAttrib(x, R_LevelsSymbol));
-			INTEGER(CAR(ans_ptr))[0] = INTEGER(x)[i];
-			ans_ptr = CDR(ans_ptr);
-		}
-		break;
 	case REALSXP:
 		for(i=0 ; i<LENGTH(x) ; i++) {
 			CAR(ans_ptr) = allocVector(REALSXP, 1);
@@ -145,7 +122,6 @@ static void listanswer(SEXP x, int recurse)
 			ans_ptr = CDR(ans_ptr);
 		}
 		break;
-#ifdef COMPLEX_DATA
 	case CPLXSXP:
 		for(i=0 ; i<LENGTH(x) ; i++) {
 			CAR(ans_ptr) = allocVector(CPLXSXP, 1);
@@ -153,7 +129,6 @@ static void listanswer(SEXP x, int recurse)
 			ans_ptr = CDR(ans_ptr);
 		}
 		break;
-#endif
 	case STRSXP:
 		for(i=0 ; i<LENGTH(x) ; i++) {
 			CAR(ans_ptr) = allocVector(STRSXP, 1);
@@ -199,7 +174,7 @@ static void stringanswer(SEXP x)
 		break;
 	default:
 		PROTECT(x = coerceVector(x, STRSXP));
-		nx = LENGTH(x);  
+		nx = LENGTH(x);
 		for (i = 0; i < nx; i++)
 			STRING(ans_ptr)[ans_length++] = STRING(x)[i];
 		UNPROTECT(1);
@@ -221,7 +196,7 @@ static void integeranswer(SEXP x)
 		}
 		break;
 	default:
-		nx = LENGTH(x);  
+		nx = LENGTH(x);
 		for (i = 0; i < nx; i++)
 			INTEGER(ans_ptr)[ans_length++] = INTEGER(x)[i];
 		break;
@@ -242,12 +217,12 @@ static void realanswer(SEXP x)
 		}
 		break;
 	case REALSXP:
-		nx = LENGTH(x);  
+		nx = LENGTH(x);
 		for (i = 0; i < nx; i++)
 			REAL(ans_ptr)[ans_length++] = REAL(x)[i];
 		break;
 	default:
-		nx = LENGTH(x);  
+		nx = LENGTH(x);
 		for (i = 0; i < nx; i++) {
 			xi = INTEGER(x)[i];
 			if(xi == NA_INTEGER)
@@ -258,7 +233,6 @@ static void realanswer(SEXP x)
 	}
 }
 
-#ifdef COMPLEX_DATA
 static void complexanswer(SEXP x)
 {
 	int i, nx, xi;
@@ -273,7 +247,7 @@ static void complexanswer(SEXP x)
 		}
 		break;
 	case REALSXP:
-		nx = LENGTH(x);  
+		nx = LENGTH(x);
 		for (i = 0; i < nx; i++) {
 			COMPLEX(ans_ptr)[ans_length].r = REAL(x)[i];
 			COMPLEX(ans_ptr)[ans_length].i = 0.0;
@@ -281,12 +255,12 @@ static void complexanswer(SEXP x)
 		}
 		break;
 	case CPLXSXP:
-		nx = LENGTH(x);  
+		nx = LENGTH(x);
 		for (i = 0; i < nx; i++)
 			COMPLEX(ans_ptr)[ans_length++] = COMPLEX(x)[i];
 		break;
 	default:
-		nx = LENGTH(x);  
+		nx = LENGTH(x);
 		for (i = 0; i < nx; i++) {
 			xi = INTEGER(x)[i];
 			if(xi == NA_INTEGER)
@@ -296,16 +270,17 @@ static void complexanswer(SEXP x)
 		break;
 	}
 }
-#endif
 
-/* 
+/*
    TagName can get either a SYMSXP, a STRSXP or a CHARSXP or NILSXP
    for either tag or base; the first part extracts the CHARSXP needed;
    the remainer acts on that
 */
 static SEXP TagName(SEXP tag, SEXP base, int i)
 {
-	SEXP ans;
+	SEXP ans, t;
+
+	PROTECT(t=mkChar(""));
 
 	switch(TYPEOF(tag)) {
 	case SYMSXP:
@@ -316,6 +291,7 @@ static SEXP TagName(SEXP tag, SEXP base, int i)
 		break;
 	case CHARSXP:
 	case NILSXP:
+		tag = t;
 		break;
 	default:
 		error("wrong tag argument to TagName\n");
@@ -355,12 +331,13 @@ static SEXP TagName(SEXP tag, SEXP base, int i)
 			sprintf(CHAR(ans), "%s.%s",CHAR(base),CHAR(tag));
 		}
 	}
+	UNPROTECT(1);
 	return ans;
 }
 
-/* 
-   since ExtractNames is done recursively and it contains a check on the 
-   names found there must be some mechanism for saying not to check to them; 
+/*
+   since ExtractNames is done recursively and it contains a check on the
+   names found there must be some mechanism for saying not to check to them;
    hence check
 */
 static void ExtractNames(SEXP args, int recurse, int check, SEXP base)
@@ -371,81 +348,76 @@ static void ExtractNames(SEXP args, int recurse, int check, SEXP base)
 	PROTECT(blank=mkChar(""));
 
 	while(args != R_NilValue) {
-		if(!isNull(CAR(args))) 
-  		    if(isVector(CAR(args))) {
-			if(!isNull(TAG(args))) {
-				switch(length(CAR(args))) {
-				case 0:
-					break;
-				case 1:
-					STRING(ans_names)[ans_nnames++] = TagName(TAG(args),base, offset);
-					break;
-				default:
-					for(i=0 ; i<length(CAR(args)) ; i++)
-						STRING(ans_names)[ans_nnames++] = TagName(TAG(args), base, i+1+offset);
-				}
-			}
-			else {
-				if(base == R_NilValue) {
-					if(isNull(s = getAttrib(CAR(args), R_NamesSymbol))) {
-						for(i=0 ; i<length(CAR(args)) ; i++)
-							STRING(ans_names)[ans_nnames++] = blank;
-					}
-					else {
-						for(i=0 ; i<length(CAR(args)) ; i++)
-							STRING(ans_names)[ans_nnames++] = STRING(s)[i];
-					}
-				}
-				else {
-					if(isNull(s = getAttrib(CAR(args), R_NamesSymbol))) {
-						for(i=0 ; i<length(CAR(args)) ; i++) 
-							STRING(ans_names)[ans_nnames++] = TagName(base,R_NilValue,i+1+offset);
-					}
-					else {
-						for(i=0 ; i<length(CAR(args)) ; i++) 
-							STRING(ans_names)[ans_nnames++] = TagName(STRING(s)[i],base,offset);
-					}
-					offset+=i;
-				}
-			}
+	  if(!isNull(CAR(args))) {
+	    if(isVector(CAR(args))) {
+		if(!isNull(TAG(args))) {
+		  switch(length(CAR(args))) {
+		  case 0:
+		    break;
+		  case 1:
+		    STRING(ans_names)[ans_nnames++] = TagName(TAG(args),base, offset);
+		    break;
+		  default:
+		    for(i=0 ; i<length(CAR(args)) ; i++)
+		      STRING(ans_names)[ans_nnames++] = TagName(TAG(args), base, i+1+offset);
+		  }
+		} else {
+		  if(base == R_NilValue) {
+		    if(isNull(s = getAttrib(CAR(args), R_NamesSymbol))) {
+		      for(i=0 ; i<length(CAR(args)) ; i++)
+			STRING(ans_names)[ans_nnames++] = blank;
+		    }
+		    else {
+		      for(i=0 ; i<length(CAR(args)) ; i++)
+			STRING(ans_names)[ans_nnames++] = STRING(s)[i];
+		    }
+		  } else {
+		    if(isNull(s = getAttrib(CAR(args), R_NamesSymbol))) {
+		      for(i=0 ; i<length(CAR(args)) ; i++)
+			STRING(ans_names)[ans_nnames++] = TagName(base,R_NilValue,i+1+offset);
+		    } else {
+		      for(i=0 ; i<length(CAR(args)) ; i++)
+			STRING(ans_names)[ans_nnames++] = TagName(STRING(s)[i],base,offset);
+		    }
+		    offset+=i;
+		  }
 		}
-		else if(isList(CAR(args))) {
-			base = TAG(args);
-			if(recurse) {
-				ExtractNames(CAR(args), recurse, 0, base);
-			}
-			else {
-				i = 1;
-				s = CAR(args);
-				while(s != R_NilValue) {
-					if( isNull(base) ) {
-						if(!isNull(TAG(s)))
-							STRING(ans_names)[ans_nnames++] = TagName(TAG(s),base,0);
-						else 
-							STRING(ans_names)[ans_nnames++] = blank;
-					}
-					else {
-						if( !isNull(TAG(s)) )
-							STRING(ans_names)[ans_nnames++] = TagName(TAG(s),base,0);
-						else
-							STRING(ans_names)[ans_nnames++] = TagName(TAG(s),base,i);
-					}
-					s = CDR(s);
-					i++;
-				}
-			}
-			base = R_NilValue;
+	    } else if(isList(CAR(args))) {
+		base = TAG(args);
+		if(recurse) {
+		  ExtractNames(CAR(args), recurse, 0, base);
+		} else {
+		  i = 1;
+		  s = CAR(args);
+		  while(s != R_NilValue) {
+		    if( isNull(base) ) {
+		      if(!isNull(TAG(s)))
+			STRING(ans_names)[ans_nnames++] = TagName(TAG(s),base,0);
+		      else
+			STRING(ans_names)[ans_nnames++] = blank;
+		    } else {
+		      if( !isNull(TAG(s)) )
+			STRING(ans_names)[ans_nnames++] = TagName(TAG(s),base,0);
+		      else
+			STRING(ans_names)[ans_nnames++] = TagName(TAG(s),base,i);
+		    }
+		    s = CDR(s);
+		    i++;
+		  }
 		}
-		else {
-			if(!isNull(TAG(args)))
-				STRING(ans_names)[ans_nnames++] = TAG(args);
-			else
-				STRING(ans_names)[ans_nnames++] = blank;
-		}
-		args = CDR(args);
-	}
+		base = R_NilValue;
+	    } else { /* neither	 Vector	 nor  List */
+		if(!isNull(TAG(args)))
+		  STRING(ans_names)[ans_nnames++] = TAG(args);
+		else
+		  STRING(ans_names)[ans_nnames++] = blank;
+	    }
+	  } /* if(! null...) */
+	  args = CDR(args);
+	} /* while */
 	if( check && ans_nnames != ans_length) {
-		printf("INTERNAL ERROR: ans_nnames = %d    ans_length = %d\n", ans_nnames, ans_length);
+		printf("INTERNAL ERROR: ans_nnames = %d	   ans_length = %d\n",
+		       ans_nnames, ans_length);
 		error("incorrect names vector length\n");
 	}
 	UNPROTECT(1);
@@ -530,7 +502,7 @@ SEXP do_c(SEXP call, SEXP op, SEXP args, SEXP env)
 		/* (perhaps a list if recursive=F) then we must */
 		/* return a list.  Otherwise, if there was a character */
 		/* mode argument we must return a character vector. */
-		/* and otherwise, we use the natural coersion. */
+		/* and otherwise, we use the natural coercion. */
 		/* The exception is "factor" vectors.  These can */
 		/* only be combined with factors of exactly the */
 		/* same "orderedness", number of levels and */
@@ -542,34 +514,12 @@ SEXP do_c(SEXP call, SEXP op, SEXP args, SEXP env)
 	else {
 		if (ans_flags & 1)
 			mode = LGLSXP;
-		if (ans_flags & 2) {
-			if(ans_flags != 2) ErrorMessage(call, ERROR_INCOMPAT_ARGS);
-			t = CDR(args);
-			while(t != R_NilValue) {
-				if(!factorsConform(CAR(args), CAR(t)))
-					ErrorMessage(call, ERROR_INCOMPAT_FACTORS);
-				t = CDR(t);
-			}
-			mode = FACTSXP;
-		}
-		if (ans_flags & 4) {
-			if(ans_flags != 4) ErrorMessage(call, ERROR_INCOMPAT_ARGS);
-                        t = CDR(args);
- 			while(t != R_NilValue){
-				if(!factorsConform(CAR(args), CAR(t)))
-					ErrorMessage(call, ERROR_INCOMPAT_FACTORS);
-				t = CDR(t);
-			}
-			mode = ORDSXP;
-		}
 		if (ans_flags & 8)
 			mode = INTSXP;
 		if (ans_flags & 16)
 			mode = REALSXP;
-#ifdef COMPLEX_DATA
 		if (ans_flags & 32)
 			mode = CPLXSXP;
-#endif
 	}
 	PROTECT(ans = allocVector(mode, ans_length));
 	ans_ptr = ans;
@@ -588,18 +538,12 @@ SEXP do_c(SEXP call, SEXP op, SEXP args, SEXP env)
 	}
 	else if(mode == STRSXP)
 		stringanswer(args);
-#ifdef COMPLEX_DATA
 	else if(mode == CPLXSXP)
 		complexanswer(args);
-#endif
 	else if(mode == REALSXP)
 		realanswer(args);
 	else
 		integeranswer(args);
-	if(mode == FACTSXP || mode == ORDSXP) {
-		LEVELS(ans) = LEVELS(CAR(args));
-		setAttrib(ans, R_LevelsSymbol, getAttrib(CAR(args), R_LevelsSymbol));
-	}
 	args = t;
 	if(ans_nnames && ans_length > 0) {
 		PROTECT(ans_names = allocVector(STRSXP, ans_length));
@@ -620,18 +564,59 @@ SEXP substituteList(SEXP, SEXP);
 
 SEXP do_bind(SEXP call, SEXP op, SEXP args, SEXP env)
 {
-	int mode;
-	SEXP t;
+	int mode=ANYSXP;	/* for -Wall; none from the ones below */
+	SEXP a, p, t;
+
+		/* First we check to see if any of the */
+		/* arguments are data frames.  If there */
+		/* are, we need to a special dispatch */
+		/* to the interpreted data.frame functions. */
+
+	mode = 0;
+	for (a = args; a != R_NilValue ; a = CDR(a)) {
+		if (isFrame(CAR(a)))
+			mode = 1;
+	}
+	if (mode) {
+		a = args;
+		t = CDR(call);
+		while (a != R_NilValue) {
+			if(t == R_NilValue)
+				errorcall(call, "corrupt data frame args!\n");
+			p = mkPROMISE(CAR(t), rho);
+			PRVALUE(p) = CAR(a);
+			CAR(a) = p;
+			t = CDR(t);
+			a = CDR(a);
+		}
+		switch(PRIMVAL(op)) {
+		    case 1:
+			op = install("cbind.data.frame");
+			break;
+		    case 2:
+			op = install("rbind.data.frame");
+			break;
+		}
+		PROTECT(op = findFun(op, env));
+		if (TYPEOF(op) != CLOSXP)
+			errorcall(call, "non closure invoked in rbind/cbind\n");
+		args = applyClosure(call, op, args, env, R_NilValue);
+		UNPROTECT(1);
+		return args;
+	}
+
+		/* There are no data frames in the argument list. */
+		/*  Perform default action */
 
 	rho = env;
 	ans_flags = 0;
 	ans_length = 0;
 	ans_nnames = 0;
 	answertype(args, 0, 0);
-
+	/* zero-extent matrices shouldn't give NULL
 	if (ans_length == 0)
 		return R_NilValue;
-
+		*/
 	if (ans_flags >= 128) {
 		if (ans_flags & 128)
 			mode = LISTSXP;
@@ -643,40 +628,17 @@ SEXP do_bind(SEXP call, SEXP op, SEXP args, SEXP env)
 	else {
 		if (ans_flags & 1)
 			mode = LGLSXP;
-		if (ans_flags & 2) {
-			if(ans_flags != 2) ErrorMessage(call, ERROR_INCOMPAT_ARGS);
-			t = CDR(args);
-			while(t != R_NilValue) {
-				if(!factorsConform(CAR(args), CAR(t)))
-					ErrorMessage(call, ERROR_INCOMPAT_FACTORS);
-				t = CDR(t);
-			}
-			mode = FACTSXP;
-		}
-		if (ans_flags & 4) {
-			if(ans_flags != 4) ErrorMessage(call, ERROR_INCOMPAT_ARGS);
-			while(t != R_NilValue) {
-				if(!factorsConform(CAR(args), CAR(t)))
-					ErrorMessage(call, ERROR_INCOMPAT_FACTORS);
-				t = CDR(t);
-			}
-			mode = ORDSXP;
-		}
 		if (ans_flags & 8)
 			mode = INTSXP;
 		if (ans_flags & 16)
 			mode = REALSXP;
-#ifdef COMPLEX_DATA
 		if (ans_flags & 32)
 			mode = CPLXSXP;
-#endif
 	}
 
 	switch(mode) {
 		case NILSXP:
 		case LGLSXP:
-		case FACTSXP:
-		case ORDSXP:
 		case INTSXP:
 		case REALSXP:
 		case CPLXSXP:
@@ -685,13 +647,14 @@ SEXP do_bind(SEXP call, SEXP op, SEXP args, SEXP env)
 		default:
 			errorcall(call, "cannot create a matrix from these types\n");
 	}
-	
+
 	switch (PRIMVAL(op)) {
 	case 1:
 		return cbind(call, args, mode);
 	case 2:
 		return rbind(call, args, mode);
 	}
+	return call;/* never used; just for -Wall */
 }
 
 
@@ -715,29 +678,30 @@ static SEXP cbind(SEXP call, SEXP args, SEXPTYPE mode)
 	mrnames = 0;
 	rows = 0;
 	cols = 0;
-	mrows = 0;
+	/* mrows = 0;*/
+	mrows = -1;
 
 		/* check conformability of matrix arguments */
 
 	n = 0;
 	for (t = args; t != R_NilValue; t = CDR(t)) {
-		if( length(CAR(t)) > 0 ) {
+		if( length(CAR(t)) >= 0 ) {
 			dims = getAttrib(CAR(t), R_DimSymbol);
 			if (length(dims) == 2) {
-				if (mrows == 0)
+				if (mrows == -1)
 					mrows = INTEGER(dims)[0];
 				else if (mrows != INTEGER(dims)[0])
 					errorcall(call, "number of rows of matrices must match (see arg %d)\n", n + 1);
 				cols += INTEGER(dims)[1];
 			}
-			else {
+			else if (length(CAR(t))>0) {
 				rows = imax(rows, length(CAR(t)));
 				cols += 1;
 			}
 		}
 		n++;
 	}
-	if (mrows != 0) rows = mrows;
+	if (mrows != -1) rows = mrows;
 
 		/* check conformability of vector arguments */
 		/* look for dimnames */
@@ -746,7 +710,7 @@ static SEXP cbind(SEXP call, SEXP args, SEXPTYPE mode)
 	warned = 0;
 	for (t = args; t != R_NilValue; t = CDR(t)) {
 		n++;
-		if (length(CAR(t)) > 0 ) {
+		if (length(CAR(t)) >= 0 ) {
 			dims = getAttrib(CAR(t), R_DimSymbol);
 			if (length(dims) == 2) {
 				dn = getAttrib(CAR(t), R_DimNamesSymbol);
@@ -757,7 +721,7 @@ static SEXP cbind(SEXP call, SEXP args, SEXPTYPE mode)
 			}
 			else {
 				k = length(CAR(t));
-				if(!warned && (k > rows || rows % k)) {
+				if(!warned && k>0 && (k > rows || rows % k)) {
 					warned = 1;
 					PROTECT(call = substituteList(call, rho));
 					warningcall(call, "number of rows of result\n\tis not a multiple of vector length (arg %d)\n", n);
@@ -777,7 +741,7 @@ static SEXP cbind(SEXP call, SEXP args, SEXPTYPE mode)
 	n = 0;
 
 	if (mode == STRSXP) {
-		for (t = args; t != R_NilValue; t = CDR(t)) { 
+		for (t = args; t != R_NilValue; t = CDR(t)) {
 			if( length(CAR(t)) > 0 ) {
 				u = CAR(t) = coerceVector(CAR(t), STRSXP);
 				k = LENGTH(u);
@@ -787,7 +751,6 @@ static SEXP cbind(SEXP call, SEXP args, SEXPTYPE mode)
 			}
 		}
 	}
-#ifdef COMPLEX_DATA
 	else if(mode == CPLXSXP) {
 		for (t = args; t != R_NilValue; t = CDR(t)) {
 			if( length(CAR(t)) > 0 ) {
@@ -799,7 +762,6 @@ static SEXP cbind(SEXP call, SEXP args, SEXPTYPE mode)
 			}
 		}
 	}
-#endif
 	else {
 		for (t = args; t != R_NilValue; t = CDR(t)) {
 			if( length(CAR(t)) > 0 ) {
@@ -823,17 +785,13 @@ static SEXP cbind(SEXP call, SEXP args, SEXPTYPE mode)
 			}
 		}
 	}
-	if(mode == FACTSXP || mode == ORDSXP) {
-		LEVELS(result) = LEVELS(CAR(args));
-		setAttrib(result, R_LevelsSymbol, getAttrib(CAR(args), R_LevelsSymbol));
-	}
 	if(have_cnames | have_rnames) {
 		PROTECT(blank = mkChar(""));
 		PROTECT(dn = allocList(2));
 		if(have_cnames) CADR(dn) = allocVector(STRSXP, cols);
 		j = 0;
 		for (t = args; t != R_NilValue; t = CDR(t)) {
-			if( length(CAR(t))>0) {
+			if( length(CAR(t))>=0) {
 				if (isMatrix(CAR(t))) {
 					u = getAttrib(CAR(t), R_DimNamesSymbol);
 					if(have_rnames && CAR(dn) == R_NilValue
@@ -848,7 +806,7 @@ static SEXP cbind(SEXP call, SEXP args, SEXPTYPE mode)
 								STRING(CADR(dn))[j++] = blank;
 					}
 				}
-				else {
+				else if (length(CAR(t))>0) {
 					u = getAttrib(CAR(t), R_NamesSymbol);
 					if(have_rnames && CAR(dn) == R_NilValue
 					   && u != R_NilValue && length(u) == rows)
@@ -886,25 +844,25 @@ static SEXP rbind(SEXP call, SEXP args, SEXPTYPE mode)
 
 		/* check conformability of matrix arguments */
 
-        n = 0;
-        for (t = args; t != R_NilValue; t = CDR(t)) {
-                if (length(CAR(t))>0) {
-                        dims = getAttrib(CAR(t), R_DimSymbol);
-                        if (length(dims) == 2) {
-                                if (mcols == 0)
-                                        mcols = INTEGER(dims)[1];
-                                else if (mcols != INTEGER(dims)[1])
-                                        errorcall(call, "number of columns of matrices must match (see arg %d)\n", n + 1);
-                                rows += INTEGER(dims)[0];
-                        }
-                        else {
-                                cols = imax(cols, length(CAR(t)));
-                                rows += 1;
-                        }
-                }
-                n++;
-        }
-        if (mcols != 0) cols = mcols;
+	n = 0;
+	for (t = args; t != R_NilValue; t = CDR(t)) {
+		if (length(CAR(t))>=0) {
+			dims = getAttrib(CAR(t), R_DimSymbol);
+			if (length(dims) == 2) {
+				if (mcols == 0)
+					mcols = INTEGER(dims)[1];
+				else if (mcols != INTEGER(dims)[1])
+					errorcall(call, "number of columns of matrices must match (see arg %d)\n", n + 1);
+				rows += INTEGER(dims)[0];
+			}
+			else if(length(CAR(t))>0){
+				cols = imax(cols, length(CAR(t)));
+				rows += 1;
+			}
+		}
+		n++;
+	}
+	if (mcols != 0) cols = mcols;
 
 		/* check conformability of vector arguments */
 		/* look for dimnames */
@@ -913,7 +871,7 @@ static SEXP rbind(SEXP call, SEXP args, SEXPTYPE mode)
 	warned = 0;
 	for (t = args; t != R_NilValue; t = CDR(t)) {
 		n++;
-		if (length(CAR(t))>0) {
+		if (length(CAR(t))>=0) {
 			dims = getAttrib(CAR(t), R_DimSymbol);
 			if (length(dims) == 2) {
 				dn = getAttrib(CAR(t), R_DimNamesSymbol);
@@ -924,7 +882,7 @@ static SEXP rbind(SEXP call, SEXP args, SEXPTYPE mode)
 			}
 			else {
 				k = length(CAR(t));
-				if(!warned && (k > cols || cols % k)) {
+				if(!warned && k>0 && (k > cols || cols % k)) {
 					warned = 1;
 					PROTECT(call = substituteList(call, rho));
 					warningcall(call, "number of columns of result\n\tnot a multiple of vector length (arg %d)\n", n);
@@ -944,7 +902,7 @@ static SEXP rbind(SEXP call, SEXP args, SEXPTYPE mode)
 	n = 0;
 	if (mode == STRSXP) {
 		for (t = args; t != R_NilValue; t = CDR(t)) {
-			if (length(CAR(t))>0) {
+			if (length(CAR(t))>=0) {
 				CAR(t) = coerceVector(CAR(t), STRSXP);
 				u = CAR(t);
 				k = LENGTH(u);
@@ -958,7 +916,6 @@ static SEXP rbind(SEXP call, SEXP args, SEXPTYPE mode)
 		UNPROTECT(1);
 		return result;
 	}
-#ifdef COMPLEX_DATA
 	else if (mode == CPLXSXP) {
 		for (t = args; t != R_NilValue; t = CDR(t)) {
 			if (length(CAR(t))>0) {
@@ -975,12 +932,19 @@ static SEXP rbind(SEXP call, SEXP args, SEXPTYPE mode)
 		UNPROTECT(1);
 		return result;
 	}
-#endif
 	for (t = args; t != R_NilValue; t = CDR(t)) {
-		if (length(CAR(t))>0) {
+		if (length(CAR(t))>=0) {
 			u = CAR(t);
 			k = LENGTH(u);
+			/*
 			mrows = (isMatrix(u)) ? nrows(u) : 1;
+			*/
+			if (isMatrix(u)) {
+			  mrows=nrows(u);
+			}
+			else {
+			  if (length(u)==0) mrows=0; else mrows=1;
+			}
 			if (TYPEOF(u) <= INTSXP) {
 				if (mode <= INTSXP) {
 					for (i = 0; i < mrows; i++)
@@ -1003,17 +967,13 @@ static SEXP rbind(SEXP call, SEXP args, SEXPTYPE mode)
 			}
 		}
 	}
-	if(mode == FACTSXP || mode == ORDSXP) {
-		LEVELS(result) = LEVELS(CAR(args));
-		setAttrib(result, R_LevelsSymbol, getAttrib(CAR(args), R_LevelsSymbol));
-	}
 	if(have_rnames | have_cnames) {
 		PROTECT(blank = mkChar(""));
 		PROTECT(dn = allocList(2));
 		if(have_rnames) CAR(dn) = allocVector(STRSXP, rows);
 		j = 0;
 		for (t = args; t != R_NilValue; t = CDR(t)) {
-			if (length(CAR(t))>0) {
+			if (length(CAR(t))>=0) {
 				if (isMatrix(CAR(t))) {
 					u = getAttrib(CAR(t), R_DimNamesSymbol);
 					if(have_cnames && CADR(dn) == R_NilValue
@@ -1030,7 +990,7 @@ static SEXP rbind(SEXP call, SEXP args, SEXPTYPE mode)
 						}
 					}
 				}
-				else {
+				else if (length(CAR(t))>0){
 					u = getAttrib(CAR(t), R_NamesSymbol);
 					if(have_cnames && CADR(dn) == R_NilValue
 					   && u != R_NilValue && length(u) == cols)
