@@ -21,6 +21,10 @@
 #include "Print.h"
 #include "names.h"
 
+#ifdef Unix
+SEXP do_getenv(SEXP, SEXP, SEXP, SEXP);
+#endif
+
 #ifdef Win32
 SEXP do_winedit(SEXP, SEXP, SEXP, SEXP);
 SEXP do_sysfile(SEXP, SEXP, SEXP, SEXP);
@@ -28,38 +32,61 @@ SEXP do_getenv(SEXP, SEXP, SEXP, SEXP);
 #endif
 
 
-/*
- *	printname  c-entry  offset  eval  arity	 pp-info  mark
+/* Table of  .Internal(.) and .Primitive(.)  R functions
+ * =====     =========	      ==========
  *
- *	Note:	We now consider eval to be made up of two digits XY.
- *		X = 1 says that this is an internal function which must
- *		be accessed with a .Internal call, any other value is
- *		accessable directly.
- *		Y = 1 says evaluate arguments before calling and Y=0 says
- *		don't evaluate.
- *	New Apr 9/96: eval is made up of three digits XYZ. X=1 says that
- *		we should switch R_Visible off (the least common situation).
- *		Y=1 says that this is an internal function (as above) and
- *		Z=1 says evaluate arguments before calling and Z=0 says
- *		don't evaluate.
+ * Each entry is a line with
  *
- * E.g:		SEXP do_cat(SEXP, SEXP, SEXP, SEXP);
+ * printname	c-entry		offset	eval	arity	pp-info		mark
+ * ---------	-------		------	----	-----	-------		----
+ *
+ * printname:	The function name in R
+ *
+ * c-entry:	The name of the corresponding C function,
+ *		actually declared in names.h.
+ *		Convention:
+ *		 - all start with "do_",
+ *		 - all return SEXP.
+ *		 - all have argument list
+ *			 (SEXP call, SEXP op, SEXP args, SEXP env)
+ *
+ * offset:	the 'op' (offset pointer) above; used for C functions
+ *		which deal with more than one R function...
+ *
+ * eval:	= XYZ (three digits)  [New Apr 9/96, before only had "YZ"].
+ *		  --- where e.g. '1' means '001'
+ *		X=1 says that we should switch R_Visible off
+ *		    (the least common situation).
+ *		Y=1 says that this is an internal function which must
+ *		    be accessed with a	.Internal(.) call, any other value is
+ *		    accessable directly and printed in R as ".Primitive(..)".
+ *		Z=1 says evaluate arguments before calling and
+ *		Z=0 says don't evaluate.
+ *
+ * arity:	How many arguments are required/allowed;  "-1"	meaning ``any''
+ *
+ * pp-info:	Deparsing Info (-> names.h )
+ *
+ * mark:	?? always = 0, i.e. completely superfluous (?). [MM]
  *
  */
 
 FUNTAB R_FunTab[] =
 {
+
 /* Language Related Constructs */
 
+/* printname	c-entry		offset	eval	arity	pp-info		mark
+ * ---------	-------		------	----	-----	-------		---- */
 {"if",		do_if,		0,	0,	-1,	PP_IF,		0},
 {"while",	do_while,	0,	0,	-1,	PP_WHILE,	0},
 {"for",		do_for,		0,	0,	-1,	PP_FOR,		0},
 {"repeat",	do_repeat,	0,	0,	-1,	PP_REPEAT,	0},
-{"break",	do_break,	CTXT_BREAK,0,	-1,	PP_BREAK,	0},
-{"next",	do_break,	CTXT_NEXT,0,	-1,	PP_NEXT,	0},
-{"return",	do_return,	0,	1,	-1,	PP_RETURN,	0},
-{"stop",	do_stop,	0,	1,	1,	PP_FUNCALL,	0},
-{"warning",	do_warning,	0,	101,	1,	PP_FUNCALL,	0},
+{"break",	do_break, CTXT_BREAK,	0,	-1,	PP_BREAK,	0},
+{"next",	do_break, CTXT_NEXT,	0,	-1,	PP_NEXT,	0},
+{"return",	do_return,	0,	0,	-1,	PP_RETURN,	0},
+{"stop",	do_stop,	0,	11,	1,	PP_FUNCALL,	0},
+{"warning",	do_warning,	0,	111,	1,	PP_FUNCALL,	0},
 {"function",	do_function,	0,	0,	-1,	PP_FUNCTION,	0},
 {"<-",		do_set,		1,	100,	-1,	PP_ASSIGN,	0},
 {"<<-",		do_set,		2,	100,	-1,	PP_ASSIGN2,	0},
@@ -81,6 +108,7 @@ FUNTAB R_FunTab[] =
 {"on.exit",	do_onexit,	0,	100,	1,	PP_FUNCALL,	0},
 {"Recall",	do_recall,	0,	10,	-1,	PP_FUNCALL,	0},
 {"delay",	do_delay,	0,	11,	2,	PP_FUNCALL,	0},
+{".Alias",	do_alias,	0,	1,	1,	PP_FUNCALL,	0},
 
 
 /* Binary Operators */
@@ -117,59 +145,58 @@ FUNTAB R_FunTab[] =
 
 /* Vectors, Matrices and Arrays */
 
-{"vector",	do_makevector,	0,	1,	2,	PP_FUNCALL,	0},
-{"factor",	do_makefactor,	0,	11,	3,	PP_FUNCALL,	0},
+/* printname	c-entry		offset	eval	arity	pp-info		mark
+ * ---------	-------		------	----	-----	-------		---- */
+{"vector",	do_makevector,	0,	11,	2,	PP_FUNCALL,	0},
 {"complex",	do_complex,	0,	11,	3,	PP_FUNCALL,	0},
 {"matrix",	do_matrix,	0,	11,	4,	PP_FUNCALL,	0},
-{"array",	do_array,	0,	1,	2,	PP_FUNCALL,	0},
+/*MM(98/4/22){"array",	do_array,0,	1,	2,	PP_FUNCALL,	0},*/
 {"length",	do_length,	0,	1,	1,	PP_FUNCALL,	0},
 {"length<-",	do_lengthgets,	0,	1,	2,	PP_FUNCALL,	0},
-{"nlevels",	do_nlevels,	0,	1,	1,	PP_FUNCALL,	0},
 {"row",		do_rowscols,	1,	11,	1,	PP_FUNCALL,	0},
 {"col",		do_rowscols,	2,	11,	1,	PP_FUNCALL,	0},
-{"c",		do_c,		0,	0,	-1,	PP_FUNCALL,	0},
+{"c",/* bind.c:*/do_c,		0,	0,	-1,	PP_FUNCALL,	0},
 {"unlist",	do_c,		1,	0,	-1,	PP_FUNCALL,	0},
 {"cbind",	do_bind,	1,	11,	-1,	PP_FUNCALL,	0},
-{"rbind",	do_bind,	2,	1,	-1,	PP_FUNCALL,	0},
-{"drop",	do_drop,	0,	1,	1,	PP_FUNCALL,	0},
+{"rbind",	do_bind,	2,	11,	-1,	PP_FUNCALL,	0},
+{"drop",	do_drop,	0,	11,	1,	PP_FUNCALL,	0},
 {"class",	do_class,	0,	1,	1,	PP_FUNCALL,	0},
 {"class<-",	do_classgets,	0,	1,	2,	PP_FUNCALL,	0},
 {"unclass",	do_unclass,	0,	1,	1,	PP_FUNCALL,	0},
 {"names",	do_names,	0,	11,	1,	PP_FUNCALL,	0},
 {"names<-",	do_namesgets,	0,	11,	2,	PP_FUNCALL,	0},
-{"dimnames",	do_dimnames,	0,	1,	1,	PP_FUNCALL,	0},
-{"dimnames<-",	do_dimnamesgets,0,	1,	2,	PP_FUNCALL,	0},
+{"dimnames",	do_dimnames,	0,	0,	1,	PP_FUNCALL,	0},
+{"dimnames<-",	do_dimnamesgets,0,	0,	2,	PP_FUNCALL,	0},
 {"all.names",	do_allnames,	0,	11,	4,	PP_FUNCALL,	0},
-{"row.names",	do_rownames,	0,	1,	1,	PP_FUNCALL,	0},
-{"dim",		do_dim,		0,	1,	1,	PP_FUNCALL,	0},
-{"dim<-",	do_dimgets,	0,	1,	2,	PP_FUNCALL,	0},
-{"levels",	do_levels,	0,	1,	1,	PP_FUNCALL,	0},
-{"levels<-",	do_levelsgets,	0,	1,	2,	PP_FUNCALL,	0},
+{"dim",		do_dim,		0,	0,	1,	PP_FUNCALL,	0},
+{"dim<-",	do_dimgets,	0,	0,	2,	PP_FUNCALL,	0},
 {"attributes",	do_attributes,	0,	1,	1,	PP_FUNCALL,	0},
 {"attributes<-",do_attributesgets,0,	1,	1,	PP_FUNCALL,	0},
 {"attr",	do_attr,	0,	1,	2,	PP_FUNCALL,	0},
 {"attr<-",	do_attrgets,	0,	0,	3,	PP_FUNCALL,	0},
-{"comment",	do_comment,	0,	1,	1,	PP_FUNCALL,	0},
-{"comment<-",	do_commentgets,	0,	1,	2,	PP_FUNCALL,	0},
+{"comment",	do_comment,	0,	11,	1,	PP_FUNCALL,	0},
+{"comment<-",	do_commentgets,	0,	11,	2,	PP_FUNCALL,	0},
 {"get",		do_get,		1,	10,	4,	PP_FUNCALL,	0},
 {"exists",	do_get,		0,	10,	4,	PP_FUNCALL,	0},
 {"assign",	do_assign,	0,	110,	4,	PP_FUNCALL,	0},
 {"remove",	do_remove,	0,	111,	3,	PP_FUNCALL,	0},
-{"duplicated",	do_duplicated,	0,	1,	1,	PP_FUNCALL,	0},
-{"unique",	do_duplicated,	1,	1,	1,	PP_FUNCALL,	0},
+{"duplicated",	do_duplicated,	0,	11,	1,	PP_FUNCALL,	0},
+{"unique",	do_duplicated,	1,	11,	1,	PP_FUNCALL,	0},
 {"match",	do_match,	0,	11,	3,	PP_FUNCALL,	0},
-{"pmatch",	do_pmatch,	0,	11,	2,	PP_FUNCALL,	0},
+{"pmatch",	do_pmatch,	0,	11,	3,	PP_FUNCALL,	0},
+{"charmatch",	do_charmatch,	0,	11,	2,	PP_FUNCALL,	0},
 {"match.call",	do_matchcall,	0,	11,	3,	PP_FUNCALL,	0},
 {"complete.cases",do_compcases,	0,	11,	1,	PP_FUNCALL,	0},
 {"attach",	do_attach,	0,	111,	3,	PP_FUNCALL,	0},
 {"detach",	do_detach,	0,	111,	1,	PP_FUNCALL,	0},
-{"search",	do_search,	0,	1,	0,	PP_FUNCALL,	0},
+{"search",	do_search,	0,	11,	0,	PP_FUNCALL,	0},
 
 
 /* Mathematical Functions */
-{"round",	do_round,	10001,	1,	1,	PP_FUNCALL,	0},
+{"round",	do_round,	10001,	11,	1,	PP_FUNCALL,	0},
 {"atan",	do_atan,	10002,	1,	1,	PP_FUNCALL,	0},
-{"log",		do_log,		10003,	1,	1,	PP_FUNCALL,	0},
+{"log",		do_log,		10003,	11,	1,	PP_FUNCALL,	0},
+{"signif",	do_signif,	10004,	11,	1,	PP_FUNCALL,	0},
 
 {"abs",		do_math1,	0,	1,	1,	PP_FUNCALL,	0},
 {"floor",	do_math1,	1,	1,	1,	PP_FUNCALL,	0},
@@ -193,25 +220,24 @@ FUNTAB R_FunTab[] =
 {"asinh",	do_math1,	34,	1,	1,	PP_FUNCALL,	0},
 {"atanh",	do_math1,	35,	1,	1,	PP_FUNCALL,	0},
 
-{"lgamma",	do_math1,	40,	1,	1,	PP_FUNCALL,	0},
-{"gamma",	do_math1,	41,	1,	1,	PP_FUNCALL,	0},
+{"lgamma",	do_math1,	40,	11,	1,	PP_FUNCALL,	0},
+{"gamma",	do_math1,	41,	11,	1,	PP_FUNCALL,	0},
 
 /* Polygamma Functions */
 
-{"digamma",	do_math1,	42,	1,	1,	PP_FUNCALL,	0},
-{"trigamma",	do_math1,	43,	1,	1,	PP_FUNCALL,	0},
-{"tetragamma",	do_math1,	44,	1,	1,	PP_FUNCALL,	0},
-{"pentagamma",	do_math1,	45,	1,	1,	PP_FUNCALL,	0},
+{"digamma",	do_math1,	42,	11,	1,	PP_FUNCALL,	0},
+{"trigamma",	do_math1,	43,	11,	1,	PP_FUNCALL,	0},
+{"tetragamma",	do_math1,	44,	11,	1,	PP_FUNCALL,	0},
+{"pentagamma",	do_math1,	45,	11,	1,	PP_FUNCALL,	0},
 
 /* Mathematical Functions of Two Variables */
 
-{"atan2",	do_math2,	0,	1,	2,	PP_FUNCALL,	0},
-{"signif",	do_math2,	1,	1,	2,	PP_FUNCALL,	0},
+{"atan2",	do_math2,	0,	11,	2,	PP_FUNCALL,	0},
 
-{"lbeta",	do_math2,	2,	1,	2,	PP_FUNCALL,	0},
-{"beta",	do_math2,	3,	1,	2,	PP_FUNCALL,	0},
-{"lchoose",	do_math2,	4,	1,	2,	PP_FUNCALL,	0},
-{"choose",	do_math2,	5,	1,	2,	PP_FUNCALL,	0},
+{"lbeta",	do_math2,	2,	11,	2,	PP_FUNCALL,	0},
+{"beta",	do_math2,	3,	11,	2,	PP_FUNCALL,	0},
+{"lchoose",	do_math2,	4,	11,	2,	PP_FUNCALL,	0},
+{"choose",	do_math2,	5,	11,	2,	PP_FUNCALL,	0},
 
 {"dchisq",	do_math2,	6,	11,	2,	PP_FUNCALL,	0},
 {"pchisq",	do_math2,	7,	11,	2,	PP_FUNCALL,	0},
@@ -234,7 +260,6 @@ FUNTAB R_FunTab[] =
 {"qt",		do_math2,	20,	11,	2,	PP_FUNCALL,	0},
 
 
-#ifdef COMPLEX_DATA
 /* Mathematical Functions of a Complex Argument */
 
 {"Re",		do_cmathfuns,	1,	1,	1,	PP_FUNCALL,	0},
@@ -243,7 +268,6 @@ FUNTAB R_FunTab[] =
 {"Arg",		do_cmathfuns,	4,	1,	1,	PP_FUNCALL,	0},
 {"Conj",	do_cmathfuns,	5,	1,	1,	PP_FUNCALL,	0},
 
-#endif
 
 /* Mathematical Functions of Three Variables */
 
@@ -295,6 +319,13 @@ FUNTAB R_FunTab[] =
 {"pnchisq",	do_math3,	35,	11,	3,	PP_FUNCALL,	0},
 {"qnchisq",	do_math3,	36,	11,	3,	PP_FUNCALL,	0},
 
+{"dnt",		do_math3,	37,	11,	3,	PP_FUNCALL,	0},
+{"pnt",		do_math3,	38,	11,	3,	PP_FUNCALL,	0},
+{"qnt",		do_math3,	39,	11,	3,	PP_FUNCALL,	0},
+
+{"dwilcox",	do_math3,	40,	11,	3,	PP_FUNCALL,	0},
+{"pwilcox",	do_math3,	41,	11,	3,	PP_FUNCALL,	0},
+{"qwilcox",	do_math3,	42,	11,	3,	PP_FUNCALL,	0},
 
 /* Mathematical Functions of Four Variables */
 
@@ -302,6 +333,17 @@ FUNTAB R_FunTab[] =
 {"phyper",	do_math4,	2,	11,	4,	PP_FUNCALL,	0},
 {"qhyper",	do_math4,	3,	11,	4,	PP_FUNCALL,	0},
 
+{"dnbeta",	do_math4,	4,	11,	4,	PP_FUNCALL,	0},
+{"pnbeta",	do_math4,	5,	11,	4,	PP_FUNCALL,	0},
+{"qnbeta",	do_math4,	6,	11,	4,	PP_FUNCALL,	0},
+
+{"dnf",		do_math4,	7,	11,	4,	PP_FUNCALL,	0},
+{"pnf",		do_math4,	8,	11,	4,	PP_FUNCALL,	0},
+{"qnf",		do_math4,	9,	11,	4,	PP_FUNCALL,	0},
+
+{"dtukey",	do_math4,	10,	11,	4,	PP_FUNCALL,	0},
+{"ptukey",	do_math4,	11,	11,	4,	PP_FUNCALL,	0},
+{"qtukey",	do_math4,	12,	11,	4,	PP_FUNCALL,	0},
 
 /* Random Numbers */
 
@@ -322,6 +364,7 @@ FUNTAB R_FunTab[] =
 {"rnorm",	do_random2,	8,	11,	3,	PP_FUNCALL,	0},
 {"runif",	do_random2,	9,	11,	3,	PP_FUNCALL,	0},
 {"rweibull",	do_random2,	10,	11,	3,	PP_FUNCALL,	0},
+{"rwilcox",	do_random2,	11,	11,	3,	PP_FUNCALL,	0},
 
 {"rhyper",	do_random3,	0,	11,	4,	PP_FUNCALL,	0},
 
@@ -330,11 +373,11 @@ FUNTAB R_FunTab[] =
 
 /* Data Summaries */
 
-{"sum",		do_summary,	0,	11,	1,	PP_FUNCALL,	0},
-{"mean",	do_summary,	1,	11,	1,	PP_FUNCALL,	0},
-{"min",		do_summary,	2,	11,	1,	PP_FUNCALL,	0},
-{"max",		do_summary,	3,	11,	1,	PP_FUNCALL,	0},
-{"prod",	do_summary,	4,	11,	1,	PP_FUNCALL,	0},
+{"sum",		do_summary,	0,	11,	-1,	PP_FUNCALL,	0},
+  /*MM{"mean",	do_summary,	1,	11,	1,	PP_FUNCALL,	0},*/
+{"min",		do_summary,	2,	11,	-1,	PP_FUNCALL,	0},
+{"max",		do_summary,	3,	11,	-1,	PP_FUNCALL,	0},
+{"prod",	do_summary,	4,	11,	-1,	PP_FUNCALL,	0},
 {"cov",		do_cov,		0,	11,	3,	PP_FUNCALL,	0},
 {"cor",		do_cov,		1,	11,	3,	PP_FUNCALL,	0},
 
@@ -345,30 +388,24 @@ FUNTAB R_FunTab[] =
 
 /* Type coercion */
 
-{"as.factor",	do_as,		FACTSXP,11,	1,	PP_FUNCALL,	0},
-{"as.unordered",do_as,		FACTSXP,11,	1,	PP_FUNCALL,	0},
-{"as.ordered",	do_as,		ORDSXP,	11,	1,	PP_FUNCALL,	0},
-{"as.vector",	do_as,		101,	11,	2,	PP_FUNCALL,	0},
-{"as.name",	do_as,		102,	11,	1,	PP_FUNCALL,	0},
-{"as.matrix.data.frame", do_asmatrixdf,	0, 11,	1,	PP_FUNCALL,	0},
-{"codes",	do_codes,	0,	1,	1,	PP_FUNCALL,	0},
-{"codes<-",	do_codesgets,	0,	1,	2,	PP_FUNCALL,	0},
+{"as.vector",	do_asvector,	0,	10,	2,	PP_FUNCALL,	0},
 {"paste",	do_paste,	0,	11,	3,	PP_FUNCALL,	0},
-{"format",	do_format,	0,	1,	-1,	PP_FUNCALL,	0},
-{"format.info",	do_formatinfo,	0,	1,	1,	PP_FUNCALL,	0},
+{"format",	do_format,	0,	11,	-1,	PP_FUNCALL,	0},
+{"format.info",	do_formatinfo,	0,	11,	1,	PP_FUNCALL,	0},
 {"cat",		do_cat,		0,	111,	6,	PP_FUNCALL,	0},
 {"call",	do_call,	0,	0,	-1,	PP_FUNCALL,	0},
-{"do.call",	do_docall,	0,	1,	2,	PP_FUNCALL,	0},
+{"do.call",	do_docall,	0,	11,	2,	PP_FUNCALL,	0},
 {"as.call",	do_ascall,	0,	1,	1,	PP_FUNCALL,	0},
 {"type.convert",do_typecvt,	1,	11,	3,	PP_FUNCALL,	0},
 
 
 /* String Manipulation */
 
-{"nchar",	do_nchar,	1,	1,	1,	PP_FUNCALL,	0},
+{"nchar",	do_nchar,	1,	11,	1,	PP_FUNCALL,	0},
 {"substr",	do_substr,	1,	11,	3,	PP_FUNCALL,	0},
 {"strsplit",	do_strsplit,	1,	11,	2,	PP_FUNCALL,	0},
 {"abbreviate",	do_abbrev,	1,	11,	3,	PP_FUNCALL,	0},
+{"make.names",	do_makenames,	0,	11,	1,	PP_FUNCALL,	0},
 {"grep",	do_grep,	1,	11,	5,	PP_FUNCALL,	0},
 {"sub",		do_gsub,	0,	11,	5,	PP_FUNCALL,	0},
 {"gsub",	do_gsub,	1,	11,	5,	PP_FUNCALL,	0},
@@ -378,8 +415,6 @@ FUNTAB R_FunTab[] =
 
 {"is.null",	do_is,		NILSXP,	1,	1,	PP_FUNCALL,	0},
 {"is.logical",	do_is,		LGLSXP,	1,	1,	PP_FUNCALL,	0},
-{"is.unordered",do_is,		FACTSXP,1,	1,	PP_FUNCALL,	0},
-{"is.ordered",	do_is,		ORDSXP,	1,	1,	PP_FUNCALL,	0},
 {"is.integer",	do_is,		INTSXP,	1,	1,	PP_FUNCALL,	0},
 {"is.real",	do_is,		REALSXP,1,	1,	PP_FUNCALL,	0},
 {"is.double",	do_is,		REALSXP,1,	1,	PP_FUNCALL,	0},
@@ -409,6 +444,9 @@ FUNTAB R_FunTab[] =
 
 {"is.vector",	do_isvector,	0,	11,	2,	PP_FUNCALL,	0},
 {"is.na",	do_isna,	0,	1,	1,	PP_FUNCALL,	0},
+{"is.nan",	do_isnan,	0,	1,	1,	PP_FUNCALL,	0},
+{"is.finite",	do_isfinite,	0,	1,	1,	PP_FUNCALL,	0},
+{"is.infinite",	do_isinfinite,	0,	1,	1,	PP_FUNCALL,	0},
 
 
 /* Miscellaneous */
@@ -416,30 +454,34 @@ FUNTAB R_FunTab[] =
 #ifdef HAVE_TIMES
 {"proc.time",	do_proctime,	0,	1,	0,	PP_FUNCALL,	0},
 #endif
-{"Version",	do_version,	0,	1,	0,	PP_FUNCALL,	0},
-{"machine",	do_machine,	0,	1,	0,	PP_FUNCALL,	0},
-{"Machine",	do_Machine,	0,	1,	0,	PP_FUNCALL,	0},
+{"Version",	do_version,	0,	11,	0,	PP_FUNCALL,	0},
+{"machine",	do_machine,	0,	11,	0,	PP_FUNCALL,	0},
+{"Machine",	do_Machine,	0,	11,	0,	PP_FUNCALL,	0},
 {"system",	do_system,	0,	11,	2,	PP_FUNCALL,	0},
+#ifdef Unix
+{"getenv",	do_getenv,	0,	11,	1,	PP_FUNCALL,	0},
+#endif
 #ifdef Win32
 {"system.file",	do_sysfile,	0,	11,	2,	PP_FUNCALL,	0},
 {"getenv",	do_getenv,	0,	11,	1,	PP_FUNCALL,	0},
 #endif
 {"parse",	do_parse,	0,	11,	4,	PP_FUNCALL,	0},
-{"save",	do_save,	0,	111,	 3,	 PP_FUNCALL,	 0},
-{"load",	do_load,	0,	111,	 1,	 PP_FUNCALL,	 0},
-{"deparse",	do_deparse,	0,	1,	2,	PP_FUNCALL,	0},
+{"save",	do_save,	0,	111,	3,	PP_FUNCALL,	0},
+{"load",	do_load,	0,	111,	1,	PP_FUNCALL,	0},
+{"deparse",	do_deparse,	0,	11,	2,	PP_FUNCALL,	0},
 {"dput",	do_dput,	0,	111,	2,	PP_FUNCALL,	0},
 {"dump",	do_dump,	0,	111,	2,	PP_FUNCALL,	0},
 {"substitute",	do_substitute,	0,	0,	-1,	PP_FUNCALL,	0},
 {"quit",	do_quit,	0,	111,	1,	PP_FUNCALL,	0},
 {"interactive",	do_interactive,	0,	0,	0,	PP_FUNCALL,	0},
-{"readline",	do_readln,	0,	1,	0,	PP_FUNCALL,	0},
+{"readline",	do_readln,	0,	11,	0,	PP_FUNCALL,	0},
 {"menu",	do_menu,	0,	11,	1,	PP_FUNCALL,	0},
 {"print.default",do_printdefault,0,	111,	5,	PP_FUNCALL,	0},
-{"prmatrix",	do_printmatrix, 0,	111,	5,	PP_FUNCALL,	0},
+{"print.atomic",do_printdefault,0,	111,	5,	PP_FUNCALL,	0},
+{"print.matrix",do_printmatrix, 0,	111,	5,	PP_FUNCALL,	0},
 {"invisible",	do_invisible,	0,	101,	1,	PP_FUNCALL,	0},
-{"gc",		do_gc,		0,	101,	0,	PP_FUNCALL,	0},
-{"gcinfo",	do_gcinfo,	0,	101,	1,	PP_FUNCALL,	0},
+{"gc",		do_gc,		0,	111,	0,	PP_FUNCALL,	0},
+{"gcinfo",	do_gcinfo,	0,	111,	1,	PP_FUNCALL,	0},
 {"rep",		do_rep,		0,	11,	2,	PP_FUNCALL,	0},
 {"list",	do_makelist,	1,	1,	-1,	PP_FUNCALL,	0},
 {"split",	do_split,	0,	11,	2,	PP_FUNCALL,	0},
@@ -451,7 +493,7 @@ FUNTAB R_FunTab[] =
 {"dyn.load",	do_dynload,	0,	111,	1,	PP_FUNCALL,	0},
 {"dyn.unload",	do_dynunload,	0,	111,	1,	PP_FUNCALL,	0},
 {"ls",		do_ls,		1,	11,	2,	PP_FUNCALL,	0},
-{"typeof",	do_typeof,	1,	1,	1,	PP_FUNCALL,	0},
+{"typeof",	do_typeof,	1,	11,	1,	PP_FUNCALL,	0},
 {"eval",	do_eval,	1,	11,	2,	PP_FUNCALL,	0},
 {"expression",	do_expression,	1,	0,	-1,	PP_FUNCALL,	0},
 {"sys.parent",	do_sys,		1,	10,	-1,	PP_FUNCALL,	0},
@@ -465,59 +507,60 @@ FUNTAB R_FunTab[] =
 {"sys.function",do_sys,		9,	10,	-1,	PP_FUNCALL,	0},
 {"sort",	do_sort,	1,	11,	1,	PP_FUNCALL,	0},
 {"psort",	do_psort,	0,	11,	2,	PP_FUNCALL,	0},
-{"order",	do_order,	0,	1,	-1,	PP_FUNCALL,	0},
-{"rank",	do_rank,	0,	1,	1,	PP_FUNCALL,	0},
+{"order",	do_order,	0,	11,	-1,	PP_FUNCALL,	0},
+{"rank",	do_rank,	0,	11,	1,	PP_FUNCALL,	0},
 {"missing",	do_missing,	1,	0,	1,	PP_FUNCALL,	0},
 {"nargs",	do_nargs,	1,	0,	0,	PP_FUNCALL,	0},
 {"scan",	do_scan,	0,	11,	10,	PP_FUNCALL,	0},
 {"count.fields",do_countfields,	0,	11,	3,	PP_FUNCALL,	0},
-{"t.default",	do_transpose,	0,	1,	1,	PP_FUNCALL,	0},
+{"t.default",	do_transpose,	0,	11,	1,	PP_FUNCALL,	0},
 {"aperm",	do_aperm,	0,	11,	3,	PP_FUNCALL,	0},
 {"builtins",	do_builtins,	0,	11,	1,	PP_FUNCALL,	0},
 {"edit",	do_edit,	0,	11,	3,	PP_FUNCALL,	0},
-{"dataentry",	do_dataentry,	0,	1,	1,	PP_FUNCALL,	0},
-{"args",	do_args,	0,	1,	1,	PP_FUNCALL,	0},
+{"dataentry",	do_dataentry,	0,	11,	1,	PP_FUNCALL,	0},
+{"args",	do_args,	0,	11,	1,	PP_FUNCALL,	0},
 {"formals",	do_formals,	0,	11,	1,	PP_FUNCALL,	0},
 {"body",	do_body,	0,	11,	1,	PP_FUNCALL,	0},
 {"globalenv",	do_globalenv,	0,	1,	0,	PP_FUNCALL,	0},
 {"environment",	do_envir,	0,	11,	1,	PP_FUNCALL,	0},
 {"environment<-",do_envirgets,	0,	1,	2,	PP_FUNCALL,	0},
 {"options",	do_options,	0,	11,	1,	PP_FUNCALL,	0},
-{"check.bounds",do_checkbounds,	0,	1,	1,	PP_FUNCALL,	0},
-{"sink",	do_sink,	0,	101,	1,	PP_FUNCALL,	0},
-{"lib.fixup",	do_libfixup,	0,	101,	2,	PP_FUNCALL,	0},
+{"check.bounds",do_checkbounds,	0,	11,	1,	PP_FUNCALL,	0},
+{"sink",	do_sink,	0,	111,	1,	PP_FUNCALL,	0},
+{"lib.fixup",	do_libfixup,	0,	111,	2,	PP_FUNCALL,	0},
 {"pos.to.env",	do_pos2env,	0,	1,	1,	PP_FUNCALL,	0},
 
-/* Data Frames */
-
-{"data.frame",	do_dataframe,	0,	11,	3,	PP_FUNCALL,	0},
-{"is.data.frame",do_is,		80,	1,	1,	PP_FUNCALL,	0},
-{"[.data.frame",do_subsetdf,	0,	0,	-1,	PP_FUNCALL,	0},
-{"[[.data.frame",do_subsetdf2,	0,	0,	-1,	PP_FUNCALL,	0},
-{"[<-.data.frame",do_subassigndf,0,	0,	-1,	PP_FUNCALL,	0},
-{"[[<-.data.frame",do_subassigndf2,0,	0,	-1,	PP_FUNCALL,	0},
-{"print.data.frame",do_printdf,	0,	101,	-1,	PP_FUNCALL,	0},
-{"any.data.frame",do_anydf,	0,	1,	-1,	PP_FUNCALL,	0},
-
 /* Complex Valued Functions */
-#ifdef COMPLEX_DATA
 {"fft",		do_fft,		0,	11,	2,	PP_FUNCALL,	0},
 {"mvfft",	do_mvfft,	0,	11,	2,	PP_FUNCALL,	0},
-{"polyroot",	do_polyroot,	0,	1,	1,	PP_FUNCALL,	0},
-#endif
+{"polyroot",	do_polyroot,	0,	11,	1,	PP_FUNCALL,	0},
 {"nextn",	do_nextn,	0,	11,	2,	PP_FUNCALL,	0},
 
+/* Device Drivers */
+
+{"X11",		do_X11,		0,	111,	4,	PP_FUNCALL,	0},
+{"PS",		do_PS,		0,	111,	9,	PP_FUNCALL,	0},
+{"PicTeX",	do_PicTeX,	0,	111,	6,	PP_FUNCALL, 	0},
+
 /* Graphics */
 
+{"dev.control",	do_devcontrol,	0,	111,	0,	PP_FUNCALL,	0},
+{"dev.copy",	do_devcopy,	0,	111,	1,	PP_FUNCALL,	0},
+{"dev.cur",	do_devcur,	0,	111,	0,	PP_FUNCALL,	0},
+/*
 {"device",	do_device,	0,	111,	3,	PP_FUNCALL,	0},
-{"dev.off",	do_devoff,	0,	101,	0,	PP_FUNCALL,	0},
+*/
+{"dev.next",	do_devnext,	0,	111,	1,	PP_FUNCALL,	0},
+{"dev.off",	do_devoff,	0,	111,	1,	PP_FUNCALL,	0},
+{"dev.prev",	do_devprev,	0,	111,	1,	PP_FUNCALL,	0},
+{"dev.set",	do_devset,	0,	111,	1,	PP_FUNCALL,	0},
 {"rgb",		do_rgb,		0,	11,	4,	PP_FUNCALL,	0},
 {"hsv",		do_hsv,		0,	11,	4,	PP_FUNCALL,	0},
-{"gray",	do_gray,	0,	1,	1,	PP_FUNCALL,	0},
-{"colors",	do_colors,	0,	1,	0,	PP_FUNCALL,	0},
+{"gray",	do_gray,	0,	11,	1,	PP_FUNCALL,	0},
+{"colors",	do_colors,	0,	11,	0,	PP_FUNCALL,	0},
 {"palette",	do_palette,	0,	11,	1,	PP_FUNCALL,	0},
 {"plot.new",	do_plot_new,	0,	111,	1,	PP_FUNCALL,	0},
-{"plot.window",	do_plot_window,	0,	101,	3,	PP_FUNCALL,	0},
+{"plot.window",	do_plot_window,	0,	111,	3,	PP_FUNCALL,	0},
 {"axis",	do_axis,	0,	111,	7,	PP_FUNCALL,	0},
 {"plot.xy",	do_plot_xy,	0,	111,	6,	PP_FUNCALL,	0},
 {"text",	do_text,	0,	111,	6,	PP_FUNCALL,	0},
@@ -528,20 +571,23 @@ FUNTAB R_FunTab[] =
 {"rect",	do_rect,	0,	111,	6,	PP_FUNCALL,	0},
 {"polygon",	do_polygon,	0,	111,	5,	PP_FUNCALL,	0},
 {"par",		do_par,		0,	11,	1,	PP_FUNCALL,	0},
-{"par2",	do_par2,	0,	11,	1,	PP_FUNCALL,	0},
 {"segments",	do_segments,	0,	111,	6,	PP_FUNCALL,	0},
 {"arrows",	do_arrows,	0,	111,	9,	PP_FUNCALL,	0},
+{"layout",	do_layout,	0,	111,	10,	PP_FUNCALL,	0},
 {"locator",	do_locator,	0,	11,	1,	PP_FUNCALL,	0},
 {"identify",	do_identify,	0,	11,	3,	PP_FUNCALL,	0},
-{"strheight",   do_strheight,   0,      11,     3, 	PP_FUNCALL, 	0},
+{"strheight",	do_strheight,	0,	11,	3,	PP_FUNCALL,	0},
 {"strwidth",	do_strwidth,	0,	11,	3,	PP_FUNCALL,	0},
 {"contour",	do_contour,	0,	11,	6,	PP_FUNCALL,	0},
 {"image",	do_image,	0,	11,	5,	PP_FUNCALL,	0},
 {"dend",	do_dend,	0,	111,	6,	PP_FUNCALL,	0},
-{"save.plot",	do_saveplot,	0,	101,	1,	PP_FUNCALL,	0},
-{"print.plot",	do_printplot,	0,	101,	0,	PP_FUNCALL,	0},
-{"text.math",	do_mathtext,	0,	100,	1,	PP_FUNCALL,	0},
+{"dend.window",	do_dendwindow,	0,	111,	6,	PP_FUNCALL,	0},
+{"replay",	do_replay,	0,	111,	0,	PP_FUNCALL,	0},
 {"erase",	do_erase,	0,	111,	1,	PP_FUNCALL,	0},
+{"dotplot",	do_dotplot,	0,	111,	1,	PP_FUNCALL,	0},
+#ifdef PERSP
+{"persp",       do_persp,       0,      111,    4,      PP_FUNCALL,	0},
+#endif
 
 /* Objects */
 {"UseMethod",	do_usemethod,	0,	 0,	-1,	PP_FUNCALL,	0},
@@ -553,11 +599,11 @@ FUNTAB R_FunTab[] =
 {"fmin",	do_fmin,	0,	11,	4,	PP_FUNCALL,	0},
 {"zeroin",	do_zeroin,	0,	11,	4,	PP_FUNCALL,	0},
 {"terms.formula",do_termsform,	0,	11,	5,	PP_FUNCALL,	0},
-{"update.formula",do_updateform,0,	1,	2,	PP_FUNCALL,	0},
+{"update.formula",do_updateform,0,	11,	2,	PP_FUNCALL,	0},
 {"model.frame",	do_modelframe,	0,	11,	5,	PP_FUNCALL,	0},
 {"model.matrix",do_modelmatrix,	0,	11,	2,	PP_FUNCALL,	0},
 
-{"D",		do_D,		0,	1,	2,	PP_FUNCALL,	0},
+{"D",		do_D,		0,	11,	2,	PP_FUNCALL,	0},
 {"deriv.default",do_deriv,	0,	11,	4,	PP_FUNCALL,	0},
 
 {NULL,		NULL,		0,	0,	0,	0,		0},
@@ -675,21 +721,25 @@ void InitNames()
 /* Returns the symbol corresponding to the string "name". */
 SEXP install(char *name)
 {
+	char buf[128];
 	SEXP sym;
 	int i;
 
 	if(*name == '\0')
 		error("attempt to use zero-length variable name\n");
+	if (strlen(name) > 127)
+		error("symbol print-name too long\n");
+	strcpy(buf, name);
 
-	i = hashpjw(name);
+	i = hashpjw(buf);
 
 		/* check to see if the symbol is already there */
 	for (sym = R_SymbolTable[i]; sym != R_NilValue; sym = CDR(sym))
-		if (strcmp(name, CHAR(PRINTNAME(CAR(sym)))) == 0)
+		if (strcmp(buf, CHAR(PRINTNAME(CAR(sym)))) == 0)
 			return (CAR(sym));
 
 		/* make a new symbol node and link it into the list */
-	sym = mkSYMSXP(mkChar(name), R_UnboundValue);
+	sym = mkSYMSXP(mkChar(buf), R_UnboundValue);
 	R_SymbolTable[i] = CONS(sym, R_SymbolTable[i]);
 	return (sym);
 }
